from enum import IntEnum, auto


class Flag(IntEnum):
    SKELETON = 0
    TEXTURE = auto()
    NORMALS = auto()
    TANGENTS = auto()
    BITANGENTS = auto()
    COLORS = auto()


class McsaFlags:
    def __init__(self):
        self._flags: dict[int, bool] = {}

    @property
    def named_dict(self) -> dict[str, bool]:
        return {Flag(key).name: value for key, value in self._flags.items()}

    def __getitem__(self, index: int) -> bool:
        return bool(self._flags.get(index, False))

    def __setitem__(self, index: int, value: bool):
        self._flags[index] = bool(value)

    def __str__(self):
        return str(self.named_dict)



from .base import ScFileException
from .basic import (
    FileBasicError,
    FileNotFound,
    FileSuffixUnsupported,
    FileSignatureInvalid,
)
from .decode import FileDecodingError, FileParsingError, FileUnsupportedError
from .mcsa import (
    McsaDecodingError,
    McsaCountsLimit,
    McsaUnsupportedVersion,
    McsaUnknownLinkCount,
)
from .ol import OlDecodingError, OlUnknownFourcc, OlInvalidFormat



from scfile.consts import OlString
from scfile.io.binary import BinaryFileIO


class OlFileIO(BinaryFileIO):
    def readfourcc(self):
        """Read xor encoded FourCC."""
        # Read string and skip last 0x00 byte
        string = self.read(OlString.SIZE)[:-1]

        # Xor byte if it's is not null
        return bytes(byte ^ OlString.XOR for byte in string if byte != OlString.NULL)



from dataclasses import dataclass, field
from typing import Dict, List

from scfile.consts import McsaModel


@dataclass
class Vector:
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0


@dataclass
class Texture:
    u: float = 0.0
    v: float = 0.0


@dataclass
class Polygon:
    a: int = 0
    b: int = 0
    c: int = 0


@dataclass
class Color:
    r: float = 1.0
    g: float = 1.0
    b: float = 1.0


@dataclass
class VertexBone:
    ids: Dict[int, int] = field(default_factory=dict)
    weights: Dict[int, float] = field(default_factory=dict)


@dataclass
class Vertex:
    position: Vector = field(default_factory=Vector)
    texture: Texture = field(default_factory=Texture)
    normals: Vector = field(default_factory=Vector)
    # color: Color = field(default_factory=Color)
    # bone: VertexBone = field(default_factory=VertexBone)


@dataclass
class Count:
    links: int = 0
    bones: int = 0
    vertices: int = 0
    polygons: int = 0


@dataclass
class Mesh:
    name: str = "name"
    material: str = "material"
    count: Count = field(default_factory=Count)
    vertices: List[Vertex] = field(default_factory=list)
    polygons: List[Polygon] = field(default_factory=list)
    bones: Dict[int, int] = field(default_factory=dict)

    @property
    def offset(self) -> int:
        return len(self.vertices)

    def resize(self) -> None:
        self.vertices = [Vertex() for _ in range(self.count.vertices)]
        self.polygons = [Polygon() for _ in range(self.count.polygons)]


@dataclass
class Bone:
    name: str = "bone"
    parent_id: int = McsaModel.ROOT_BONE_ID
    position: Vector = field(default_factory=Vector)
    rotation: Vector = field(default_factory=Vector)


@dataclass
class Skeleton:
    bones: List[Bone] = field(default_factory=list)


@dataclass
class Flags:
    texture: bool = False
    normals: bool = False


@dataclass
class Scale:
    position: float = 1.0
    texture: float = 1.0
    normals: float = 1.0
    weight: float = 1.0


@dataclass
class Local:
    axis: Vector = field(default_factory=Vector)
    center: Vector = field(default_factory=Vector)


@dataclass
class Model:
    meshes: List[Mesh] = field(default_factory=list)
    # skeleton: Skeleton = field(default_factory=Skeleton)
    flags: Flags = field(default_factory=Flags)
    scale: Scale = field(default_factory=Scale)
    # local: Local = field(default_factory=Local)

    def ensure_unique_names(self):
        """Updates meshes names, excluding repetitions."""
        seen_names: set[str] = set()

        for mesh in self.meshes:
            name = mesh.name or "noname"

            base_name, count = name, 2
            unique_name = f"{base_name}"

            while unique_name in seen_names:
                unique_name = f"{base_name}_{count}"
                count += 1

            mesh.name = unique_name
            seen_names.add(unique_name)



from .enums import FileSuffix


class Signature:
    """
    Big-Endian Source File Signature.
    Integer.
    """

    MIC = 0x894D4943
    OL = 0x0A9523FD
    MCSA = 0x4D435341


class Magic:
    """
    Big-Endian Output File Signature.
    List[Integer].
    """

    PNG = [0x89, 0x50, 0x4E, 0x47]
    DDS = [0x44, 0x44, 0x53, 0x20]


class Factor:
    # int range + 1
    I8 = 0x80
    U8 = 0x100
    I16 = 0x8000
    U16 = 0x10000


class OlString:
    SIZE = 17  # 16 bytes + last 0x00 byte
    XOR = ord("g")
    NULL = ord("G")


class McsaModel:
    ROOT_BONE_ID = -1
    # I dont know specific limit
    # This is done for case when file was read incorrectly
    # So as not to overflow memory
    # ? Trees and new year toys have hundreds of thousands of vertices...
    COUNT_LIMIT = 0x100000


class McsaSize:
    # Number of elements in each structure
    POSITION = 4
    TEXTURE = 2
    NORMALS = 4
    POLYGONS = 3
    COLOR = 4


# Files suffixes that can be converted
SUPPORTED_SUFFIXES = {FileSuffix.MIC, FileSuffix.OL, FileSuffix.MCSA}

# Dds cubemap faces count
# +x, -x, +y, -y, +z, -z
CUBEMAP_FACES = 6



import numpy as np


class RGBA8Converter:
    channels = 4
    dtype = np.uint8

    def __init__(self, image: bytes, width: int, height: int):
        self.image = image
        self.width = width
        self.height = height
        self.array = self.generate_image_array()

    def generate_image_array(self):
        array = np.frombuffer(self.image, dtype=self.dtype)
        array = array.reshape((self.height, self.width, self.channels))
        return array

    def invert(self) -> bytes:
        inverted = 255 - self.array[:, :, :3]
        inverted = np.concatenate((inverted, self.array[:, :, 3:]), axis=2)
        return inverted.astype(np.uint8).tobytes()



from pathlib import Path
from typing import Any, Optional

from scfile import exceptions as exc
from scfile.enums import FileSuffix
from scfile.file.dds.encoder import DdsEncoder
from scfile.file.decoder import FileDecoder
from scfile.file.encoder import FileEncoder
from scfile.file.mcsa.decoder import McsaDecoder
from scfile.file.mic.decoder import MicDecoder
from scfile.file.obj.encoder import ObjEncoder
from scfile.file.ol.decoder import OlDecoder
from scfile.file.png.encoder import PngEncoder
from scfile.types import PathLike


def mcsa_to_obj(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting `.mcsa` file to `.obj`.

    Args:
        source: Full path to `.mcsa` file.
        output (optional): Full path to output `.obj` file.
        Defaults to source path with new suffix.

    Example:
        `mcsa_to_obj("C:/file.mcsa", "C:/file.obj")`
    """

    _convert(source, output, McsaDecoder, ObjEncoder, FileSuffix.OBJ)


def mic_to_png(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting `.mic` file to `.png`.

    Args:
        source: Full path to `.mic` file.
        output (optional): Full path to output `.png` file.
        Defaults to source path with new suffix.

    Example:
        `mic_to_png("C:/file.mic", "C:/file.png")`
    """

    _convert(source, output, MicDecoder, PngEncoder, FileSuffix.PNG)


def ol_to_dds(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting `.ol` file to `.dds`.

    Args:
        source: Full path to `.ol` file.
        output (optional): Full path to output `.dds` file.
        Defaults to source path with new suffix.

    Example:
        `ol_to_dds("C:/file.ol", "C:/file.dds")`
    """

    _convert(source, output, OlDecoder, DdsEncoder, FileSuffix.DDS)


def auto(source: PathLike, output: Optional[PathLike] = None):
    """
    Automatically determines which format convert to.

    Args:
        source: Full path to encrypted file.
        output (optional): Full path to output decrypted file.
        Defaults to source path with new suffix.

    Raises:
        FileSuffixUnsupported - if source suffix not in `.mic`, `.ol`, `.mcsa`.

    Example:
        `auto("C:/file.mic", "C:/file.png")`
    """

    path = Path(source)

    match path.suffix.lstrip("."):
        case FileSuffix.MIC:
            mic_to_png(source, output)

        case FileSuffix.OL:
            ol_to_dds(source, output)

        case FileSuffix.MCSA:
            mcsa_to_obj(source, output)

        case _:
            raise exc.FileSuffixUnsupported(path)


def _convert(
    source: PathLike,
    output: Optional[PathLike],
    decoder: type[FileDecoder[Any, Any]],
    encoder: type[FileEncoder[Any]],
    new_suffix: str,
):
    src = Path(source)
    new_src = src.with_suffix(f".{new_suffix}")

    dest = Path(output) if output else new_src

    # Check that file exists
    if not src.exists() or not src.is_file():
        raise exc.FileNotFound(src)

    # If destination is directory save file in this directory
    if dest.is_dir():
        dest = Path(dest, new_src.name)

    # Convert and save file
    with decoder(src) as dec:
        dec.convert_to(encoder).save(dest)



from abc import ABC
from dataclasses import dataclass

from scfile.utils.model import Model


class FileData(ABC):
    pass


@dataclass
class ModelData(FileData):
    model: Model


@dataclass
class TextureData(FileData):
    width: int
    height: int
    linear_size: int
    fourcc: bytes
    image: bytes


@dataclass
class ImageData(FileData):
    image: bytes
